function R = crsLiveTrackHIDcomm(deviceNumber,mode,type)
% This function handles the data acquisition for LiveTrack by using the
% Psychtoolbox PsychHID function.
%
% 'deviceNumber' must be the number of the LiveTrack device. You can find it
% with the LiveTrackHIDGetDeviceNumber function.
%
% 'mode' must be either 'begin-raw', 'begin-cal', 'continue',
% 'continue-returnlast', 'continue-returnnext', 'return all' or 'end'.
%
% 'begin-raw' tells LiveTrack to start sending raw data packets
%
% 'begin-cal' tells LiveTrack to start sending calibrated data packets
%
% 'continue' reads and empties the data packets from the HID buffer. The
% packets are added to a MATLAB variable containing packets from previous
% calls to 'continue'. 
%
% 'continue-returnlast' does the same as 'continue' but also returns the 
% newest data packet in the buffer.
%
% 'continue-returnnext' does the same as 'continue' but also returns the 
% next data packet. NB. Is will wait until the next packet arrives.
%
% 'return all' does the same as 'continue' except all data packets from
% this and previous calls to 'continue' are returned and the MATLAB
% variable is emptied.
%
% 'clear buffer' does the same as 'continue' except all data packets from
% this and previous calls to 'continue' are deleted and the MATLAB
% variable is emptied.
%
% 'end' tells LiveTrack to stop sending data packets.
%
% 'type' is the LiveTrack type: must be either 'LiveTrack-AV', 
% 'LiveTrack-FM' or 'LiveTrack-AP'.
%
% On binocular systems (LiveTrack FM and AP) both left and right eye data 
% is returned. On monocular systems (based on LiveTrack AV) only data from
% one eye is returned. By default this is returned as left eye data.
%
% The following data in contained in the structure returned for raw data
% 
% For the left eye:
% LPupilMa - the semi major axis of the pupil fit
% LPupilMi - the semi minor axis of the pupil fit
% LPupilX  - the horizontal coordinate of the centre of the pupil fit
% LPupilY  - the vertical coordinate of the centre of the pupil fit
% LGlint1X - the horizontal coordinate of the centre of the first glint
% LGlint1Y - the vertical coordinate of the centre of the first glint
% LGlint2X - the horizontal coordinate of the centre of the second glint
% LGlint2Y - the vertical coordinate of the centre of the second glint
%
% and for the right eye:
% RPupilMa - the semi major axis of the pupil fit
% RPupilMi - the semi minor axis of the pupil fit
% RPupilX  - the horizontal coordinate of the centre of the pupil fit
% RPupilY  - the vertical coordinate of the centre of the pupil fit
% RGlint1X - the horizontal coordinate of the centre of the first glint
% RGlint1Y - the vertical coordinate of the centre of the first glint
% RGlint2X - the horizontal coordinate of the centre of the second glint
% RGlint2Y - the vertical coordinate of the centre of the second glint
%
% The second glint coordinates are used for two-glint systems (FM and AP),
% for one-glint systems (AV) they will just contain zeroes.

% History:
% 2016/09 JT Created
% 2016/10 JT Changed to separete the two samples that each HID packet from
% LiveTrack AV contains. The output format will then be the same for all
% platforms


persistent buffer

if nargin<2,
    error('You must select a device number and a operating mode for this function. See the help.');
end

if strcmpi(mode,'begin-raw'),
    % Tell LiveTrack to start sending data packets with a packet beginning
    % with 104
    err = PsychHID('SetReport', deviceNumber, 2, 0, uint8([104 zeros(1,63)]));
    if err.n
        fprintf('\nPsychHID: SetReport error 0x%s. %s: %s\n',hexstr(err.n),err.name,err.description);
    else
        fprintf('\nLiveTrack: Started sending raw data.\n');
    end
    % Tell PsychHID to continously receive and store HID reports in its
    % buffer
    receiveReports(deviceNumber);
    R = [];
elseif strcmpi(mode,'begin-cal'),
    % Tell LiveTrack to start sending calibrated data packets with a
    % packet beginning with 105
    err = PsychHID('SetReport', deviceNumber, 2, 0, uint8([105 zeros(1,63)]));
    if err.n
        fprintf('\nPsychHID: SetReport error 0x%s. %s: %s\n',hexstr(err.n),err.name,err.description);
    else
        fprintf('\nLiveTrack: Started sending calibrated data.\n');
    end
    % Tell PsychHID to receive and store HID reports in its buffer
    receiveReports(deviceNumber);
    R = [];
elseif  strcmpi(mode,'continue'),
    % Tell PsychHID to receive and store HID reports in its buffer
    receiveReports(deviceNumber);

    % Get all the reports from the PsychHID buffer
    [reports,err]=PsychHID('GiveMeReports',deviceNumber);
    if err.n
        fprintf('\nPsychHID: GiveMeReports error 0x%s. %s: %s\n',hexstr(err.n),err.name,err.description);
    end
    % Store the data packets in our persistent MATLAB buffer
    buffer = [buffer reports];
    R = [];
elseif  strcmpi(mode,'continue-returnlast'),
    % Update the buffer with the latest data from the USB buffer
    crsLiveTrackHIDcomm(deviceNumber,'continue');
    % Make a structure with the data from the data in the last entry of
    % the buffer
    R = LiveTrackHID2struct(buffer(end),type);
    % On LiveTrack AV one buffer sample will contain two samples. In this
    % case, one return the most recent of the two samples
    R = R(end);
elseif  strcmpi(mode,'continue-returnnext'),
    % If the buffer is empty keep updating till it is not
    if isempty(buffer),
        while isempty(buffer),
            crsLiveTrackHIDcomm(deviceNumber,'continue');
        end
    % If the buffer is not empty keep updating till the length has 
    % increased  
    else
        buffer_length = length(buffer);
        while length(buffer)>buffer_length,
            crsLiveTrackHIDcomm(deviceNumber,'continue');
        end
    end
    % Make a structure with the data from the data in the last entry of
    % the buffer
    R = LiveTrackHID2struct(buffer(end),type);
    % On LiveTrack AV one buffer sample will contain two samples. In this
    % case, one return the most recent of the two samples
    R = R(end);
elseif  strcmpi(mode,'return all'),
    % Update the buffer with the latest data from the USB buffer
    crsLiveTrackHIDcomm(deviceNumber,'continue');
    % Make a structure with the data of all the samples in the buffer
    R = LiveTrackHID2struct(buffer,type);
    % Clear the buffer
    clear buffer
elseif  strcmpi(mode,'clear buffer'),
    % Update the buffer with the latest data from the USB buffer
    crsLiveTrackHIDcomm(deviceNumber,'continue');
    % Clear the buffer
    clear buffer
    % Return an empty variable
    R=[];
elseif  strcmpi(mode,'end'),
    % Tell LiveTrack to stop sending data packets with a packet beginning
    % with 102
    err = PsychHID('SetReport', deviceNumber, 2, 0, uint8([102 zeros(1,63)]));
    if err.n
        fprintf('\nPsychHID: SetReport error 0x%s. %s: %s\n',hexstr(err.n),err.name,err.description);
    else
        fprintf('\nLiveTrack: Stopped sending raw data.\n');
    end
    % Stop receiving and saving reports from the specified USB HID device
    err = PsychHID('ReceiveReportsStop',deviceNumber);
    if err.n
        fprintf('\nPsychHID: ReceiveReportsStop error 0x%s. %s: %s\n',hexstr(err.n),err.name,err.description);
    end
    R = [];
else
   error('mode must be either "begin-raw", "begin-cal", "continue", "continue-returnlast", "continue-returnnext", "return all", "clear buffer" or "end"');
end


function receiveReports(deviceNumber)
% Read and empty the data packets from the HID buffer (small) and store 
% them in the larger PsychHID buffer (default size = 10000 reports)
err = PsychHID('ReceiveReports',deviceNumber);
if err.n
    fprintf('\nPsychHID: ReceiveReports error 0x%s. %s: %s\n',hexstr(err.n),err.name,err.description);
end


function R = LiveTrackHID2struct(reports,type)
% LiveTrackHID2struct converts report packets with raw or calibrated
% data into a structure array.

if isempty(reports)
    fprintf('No reports found\n');
    R = [];
    return
end

LFieldNames = {'LPupilMa','LPupilMi','LPupilX','LPupilY',...
        'LGlint1X','LGlint1Y','LGlint2X','LGlint2Y'};

RFieldNames = {'RPupilMa','RPupilMi','RPupilX','RPupilY',...
        'RGlint1X','RGlint1Y','RGlint2X','RGlint2Y'};

LFieldNamesCal = {'LPupilMa','LPupilMi','LGazeX','LGazeY',...
    'LGazeZ','LAzimuth','LElevation','LLongitude','LLatitude'};

RFieldNamesCal = {'RPupilMa','RPupilMi','RGazeX','RGazeY',...
    'RGazeZ','RAzimuth','RElevation','RLongitude','RLatitude'};


noOfReports = 1;

% If input is a structure, more than one data packet is present
if isstruct(reports), 

    noOfReports = length(reports);

    % pre allocate
    R(length(reports)).PsychHIDtime = 0;
    R(length(reports)).LTracked = 0;
    R(length(reports)).RTracked = 0;
    R(length(reports)).timeStamp = 0;
    R(length(reports)).PsychHIDtime = 0;
    R(length(reports)).PsychHIDtime = 0;
    R(length(reports)).marker = 0;

end

% Run through all reports
for k=1:noOfReports,
    % If monocular device type, each packet will contain two data samples
    % from the same eye. Treat both samples as left eye in that case
    if ~isempty(strfind(type,'AV')), 
        k2 = (k-1)*2+1; 
    % If binocular device type (FM or AP), each packet will contain one
    % sample from left eye and one sample from right eye
    else 
        k2 = k;
    end

    % Put the current report into the variable "r"
    if isstruct(reports),
        R(k2).PsychHIDtime = reports(k).time;
        r = double(reports(k).report);
    else
        r = double(reports);
    end

    % Marker/trigger
    R(k2).marker = r(5);

    % Decode the timestamp (frame count)
    R(k2).timeStamp = r(9)+r(10)*256+r(11)*256^2+r(12)*256^3+r(13)*256^4+r(14)*256^5+r(15)*256^6+r(16)*256^7;

    % Check the "tracked flag" for the first sample in the packet
    S1Flag = dec2bin(r(17)+r(18)*256,16);
    
    % Set the "tracked flag" for the first sample to the left eye (the 
    % field "LTracked")
    if strcmp(S1Flag(11:12),'01'), % set to look for 1 glint (LiveTrack AV)
        R(k2).LTracked = strcmp(S1Flag(13:16),'0111');
    elseif strcmp(S1Flag(11:12),'10'), % set to look for 2 glints (LiveTrack FM)
        R(k2).LTracked = strcmp(S1Flag(13:16),'1111');
    else
        if strcmp(S1Flag(16),'1'), % tracking is enabled
            error('looking for an undefined amount of glints!');
        else
            R(k2).LTracked = false;
        end
    end

    % Check the "tracked flag" for the second sample in the packet
    S2Flag = dec2bin(r(41)+r(42)*256,16);
    
    % If monocular device type, set the "tracked flag" for the second
    % sample eye to be the "tracket flag" for the next frame/samples left
    % eye
    if ~isempty(strfind(type,'AV')), 
        R(k2+1).PsychHIDtime = reports(k).time;
        if strcmp(S2Flag(11:12),'01'), % set to look for 1 glint (LiveTrack AV)
            R(k2+1).LTracked = strcmp(S2Flag(13:16),'0111');
        elseif strcmp(S2Flag(11:12),'10'), % set to look for 2 glints (LiveTrack FM)
            R(k2+1).LTracked = strcmp(S2Flag(13:16),'1111');
        else
            if strcmp(S2Flag(16),'1'), % tracking is enabled
                error('looking for an undefined amount of glints!');
            else
                R(k2+1).LTracked = false;
            end
        end
        R(k2).RTracked = false;
        R(k2+1).RTracked = false;
        R(k2+1).marker = r(5);
        R(k2+1).timeStamp = R(k2).timeStamp+1;
    else % elseif binocular
        if strcmp(S2Flag(11:12),'01'), % set to look for 1 glint (LiveTrack AV)
            R(k2).RTracked = strcmp(S2Flag(13:16),'0111');
        elseif strcmp(S2Flag(11:12),'10'), % set to look for 2 glints (LiveTrack FM)
            R(k2).RTracked = strcmp(S2Flag(13:16),'1111');
        else
            if strcmp(S2Flag(16),'1'), % tracking is enabled
                error('looking for an undefined amount of glints!');
            else
                R(k2).RTracked = false;
            end
        end
    end

    r=uint16(r);

    if r(1)==200,  % if raw data     

        % S1 scaling factor
        S1Scaling = double(r(23)+r(24)*256);

        % S2 scaling factor
        S2Scaling = double(r(47)+r(48)*256);

        c=0;
        for i=25:2:39,
            c=c+1;
            R(k2).(LFieldNames{c}) = double(typecast(r(i)+r(i+1)*256,'int16'))/S1Scaling;
        end

        c=0;
        for i=49:2:63,
            c=c+1;
            if ~isempty(strfind(type,'AV')), % if monocular
                R(k2+1).(LFieldNames{c}) = double(typecast(r(i)+r(i+1)*256,'int16'))/S2Scaling;
                R(k2).(RFieldNames{c}) = 0;
                R(k2+1).(RFieldNames{c}) = 0;
            else % elseif binocular    
                R(k2).(RFieldNames{c}) = double(typecast(r(i)+r(i+1)*256,'int16'))/S2Scaling;
            end
        end
    elseif r(1)==201,  % if calibrated data - always scaled by 32
        c=0;
        for i=23:2:39,
            c=c+1;
            R(k2).(LFieldNamesCal{c}) = double(typecast(r(i)+r(i+1)*256,'int16'))/32;
        end

        c=0;
        for i=47:2:63,
            c=c+1;
            if ~isempty(strfind(type,'AV')), % if monocular
                R(k2+1).(LFieldNamesCal{c}) = double(typecast(r(i)+r(i+1)*256,'int16'))/32;
                R(k2).(RFieldNamesCal{c}) = 0;
                R(k2+1).(RFieldNamesCal{c}) = 0;
            else % elseif binocular    
                R(k2).(RFieldNamesCal{c}) = double(typecast(r(i)+r(i+1)*256,'int16'))/32;
            end

        end

    end


end



